#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import numbers
import re
import types
from collections import OrderedDict
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import numpy

import awkward.array.base
import awkward.type
import awkward.util

class Table(awkward.array.base.AwkwardArray):
    """
    Table
    """

    ##################### class Row

    class Row(awkward.util.NDArrayOperatorsMixin):
        """
        Table.Row
        """

        __slots__ = ["_table", "_index"]

        def __init__(self, table, index):
            self._table = table
            self._index = index

        def __repr__(self):
            if self._table.istuple:
                return "({0})".format(", ".join(str(self[n]) for n in self._table._contents))
            elif getattr(self._table, "_showdict", False):
                return "<{0} {{{1}}}>".format(self._table._rowname, ", ".join("{0}: {1}".format(repr(n), str(self[n])) for n in self._table._contents))
            else:
                return "<{0} {1}>".format(self._table._rowname, self._index + self._table.rowstart)

        def __contains__(self, name):
            return name in self._table._contents

        def tolist(self):
            if self._table.istuple:
                return tuple(self._table._try_tolist(self[n]) for n in self._table._contents)
            else:
                return dict((n, self._table._try_tolist(x[self._index])) for n, x in self._table._contents.items())

        def __len__(self):
            if self._table.rowname == 'tuple':
                i = 0
                while str(i) in self._table._contents:
                    i += 1
                return i
            else:
                return len(self._table._contents)

        def __iter__(self, checkiter=True):
            if checkiter:
                self._table._checkiter()
            if self._table.rowname == 'tuple':
                i = 0
                while str(i) in self._table._contents:
                    yield self._table._contents[str(i)][self._index]
                    i += 1
            else:
                for i in self._table._contents:
                    yield i

        def __getitem__(self, where):
            if isinstance(where, awkward.util.string):
                try:
                    return self._table._contents[where][self._index]
                except KeyError:
                    raise ValueError("no column named {0}".format(repr(where)))

            elif self._util_isstringslice(where):
                contents = OrderedDict()
                for n in where:
                    try:
                        contents[n] = self._table._contents[n]
                    except KeyError:
                        raise ValueError("no column named {0}".format(repr(n)))
                table = self._table.copy(contents=contents)
                return table.Row(table, self._index)

            else:
                index = self._index
                if not isinstance(index, tuple):
                    index = (index,)
                if not isinstance(where, tuple):
                    where = (where,)
                return self._table.Row(table, index + where)

        def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
            if "out" in kwargs:
                raise NotImplementedError("in-place operations not supported")

            if method != "__call__":
                return NotImplemented

            torow = not any(not isinstance(x, Table.Row) and isinstance(x, Iterable) for x in inputs)

            inputs = list(inputs)
            for i in range(len(inputs)):
                if isinstance(inputs[i], Table.Row):
                    inputs[i] = inputs[i]._table[inputs[i]._index : inputs[i]._index + 1]

            result = getattr(ufunc, method)(*inputs, **kwargs)

            if torow:
                if isinstance(result, tuple):
                    out = []
                    for x in result:
                        if isinstance(x, Table):
                            out.append(awkward.array.objects.Methods.maybemixin(type(x), self._table.Table.Row)(x, 0))
                            out[-1]._table._showdict = True
                        else:
                            out.append(x)
                    return tuple(out)
                elif method == "at":
                    return None
                else:
                    out = awkward.array.objects.Methods.maybemixin(type(result), self._table.Table.Row)(result, 0)
                    out._table._showdict = True
                    return out

            else:
                return result

        def __eq__(self, other):
            if not isinstance(other, Table.Row):
                return False
            elif self._table is other._table and self._index == other._index:
                return True
            else:
                return set(self._table._contents) == set(other._table._contents) and all(self._table._contents[n][self._index] == other._table._contents[n][other._index] for n in self._table._contents)

        def __ne__(self, other):
            return not self.__eq__(other)

        def __getattr__(self, where):
            if where in super(Table.Row, self).__dir__():
                return super(Table.Row, self).__getattribute__(where)
            else:
                if where in self.columns:
                    try:
                        return self[where]
                    except Exception as err:
                        raise AttributeError("while trying to get column {0}, an exception occurred:\n{1}: {2}".format(repr(where), type(err), str(err)))
                else:
                    raise AttributeError("no column named {0}".format(repr(where)))

        def __dir__(self):
            return sorted(set(super(Table.Row, self).__dir__() + [x for x in self.columns if self._dir_pattern.match(x) and not keyword.iskeyword(x)]))
        _dir_pattern = re.compile(r"^[a-zA-Z_]\w*$")

        @property
        def columns(self):
            return self._table.columns

        @property
        def i0(self):
            return self["0"]

        @property
        def i1(self):
            return self["1"]

        @property
        def i2(self):
            return self["2"]

        @property
        def i3(self):
            return self["3"]

        @property
        def i4(self):
            return self["4"]

        @property
        def i5(self):
            return self["5"]

        @property
        def i6(self):
            return self["6"]

        @property
        def i7(self):
            return self["7"]

        @property
        def i8(self):
            return self["8"]

        @property
        def i9(self):
            return self["9"]

    ##################### class Table

    def __init__(self, columns1={}, *columns2, **columns3):
        self._view = None
        self._base = None
        self.rowname = "Row" if isinstance(columns1, dict) or len(columns3) > 0 else "tuple"
        self.rowstart = None
        self._contents = OrderedDict()

        seen = set()
        if isinstance(columns1, dict):
            for n, x in columns1.items():
                if n in seen:
                    raise ValueError("column {0} occurs more than once".format(repr(n)))
                seen.add(n)

                self[n] = x

            if len(columns2) != 0:
                raise TypeError("only one positional argument when first argument is a dict")

        else:
            self["0"] = columns1
            for i, x in enumerate(columns2):
                self[str(i + 1)] = x

        seen.update(self._contents)

        for n, x in columns3.items():
            if n in seen:
                raise ValueError("column {0} occurs more than once".format(repr(n)))
            seen.add(n)

            self[n] = x

    def tolist(self):
        return list(x.tolist() for x in self)

    @classmethod
    def named(cls, rowname, columns1={}, *columns2, **columns3):
        out = cls(columns1, *columns2, **columns3)
        out.rowname = rowname
        return out

    @property
    def rowname(self):
        return self._rowname

    @rowname.setter
    def rowname(self, value):
        if self.check_prop_valid:
            if not isinstance(value, awkward.util.string):
                raise TypeError("rowname must be a string")
        self._rowname = value

    def _util_rowname(self, seen):
        return self._rowname

    @property
    def rowstart(self):
        if self._rowstart is not None:
            return self._rowstart
        elif self._base is not None:
            return self._base.rowstart
        else:
            return 0

    @rowstart.setter
    def rowstart(self, value):
        if self.check_prop_valid:
            if value is not None and not isinstance(value, (numbers.Integral, numpy.integer)):
                raise TypeError("rowstart must be None or an integer")
        self._rowstart = value

    @classmethod
    def fromrec(cls, recarray):
        if not isinstance(recarray, cls.numpy.ndarray) or recarray.dtype.names is None:
            raise TypeError("recarray must be a Numpy structured array")
        out = cls()
        for n in recarray.dtype.names:
            out[n] = recarray[n]
        return out

    @classmethod
    def frompairs(cls, pairs, rowstart=0):
        out = cls()
        for n, x in pairs:
            out[n] = x
        out._rowstart = rowstart
        return out

    @classmethod
    def fromview(cls, view, base):
        if view is None:
            return base

        elif isinstance(view, tuple) and len(view) == 3 and all(cls._util_isinteger(x) for x in view):
            start, step, length = view
            out = base.copy()
            out._view = int(start), int(step), int(length)
            out._base = base
            out._rowstart = None
            return out

        elif isinstance(view, cls.numpy.ndarray) and cls._util_isintegertype(view.dtype.type):
            out = base.copy()
            out._view = view
            out._base = base
            out._rowstart = None
            return out

        else:
            raise TypeError("view must be None, a 3-tuple of integers, or a Numpy array of integers")

    def copy(self, contents=None):
        out = self.__class__.__new__(self.__class__)
        out._view = self._view
        out._base = self._base
        out._rowstart = self._rowstart
        out._rowname = self._rowname
        out._contents = self._contents
        if contents is not None and isinstance(contents, dict):
            out._contents = OrderedDict(contents.items())
        elif contents is not None:
            out._contents = OrderedDict(contents)
        else:
            out._contents = OrderedDict(self._contents.items())
        return out

    def deepcopy(self, contents=None):
        out = self.copy(contents=contents)
        out._contents = OrderedDict([(n, self._util_deepcopy(x[out._index()])) for n, x in out._contents.items()])
        out._view = None
        out._base = None
        out._rowstart = None
        return out

    def empty_like(self, **overrides):
        out = self.__class__.__new__(self.__class__)
        out._view = None
        out._base = None
        out._rowstart = None
        out._rowname = self._rowname
        out._contents = OrderedDict()
        return out

    def zeros_like(self, **overrides):
        out = self.empty_like(**overrides)
        for n, x in self._contents.items():
            if isinstance(x, self.numpy.ndarray):
                out[n] = self.numpy.zeros_like(x)
            else:
                out[n] = x.zeros_like(**overrides)
        return out

    def ones_like(self, **overrides):
        out = self.empty_like(**overrides)
        for n, x in self._contents.items():
            if isinstance(x, self.numpy.ndarray):
                out[n] = self.numpy.ones_like(x)
            else:
                out[n] = x.ones_like(**overrides)
        return out

    def __awkward_serialize__(self, serializer):
        self._valid()
        out = serializer.encode_call(
            ["awkward", "Table", "frompairs"],
            {"pairs": [
                [n, serializer(x, "Table.contents")]
                for n, x in self._contents.items()
            ]},
            {"json": self.rowstart}
        )
        if isinstance(self._view, tuple):
            start, step, length = self._view
            out = serializer.encode_call(
                ["awkward", "Table", "fromview"],
                {"tuple": [
                    {"json": start},
                    {"json": step},
                    {"json": length},
                ]},
                out,
            )

        elif isinstance(self._view, self.numpy.ndarray):
            out = serializer.encode_call(
                ["awkward", "Table", "fromview"],
                serializer(self._view, "Table.view"),
                out
            )

        return out

    @property
    def base(self):
        return self._base

    @property
    def contents(self):
        return self._contents

    @contents.setter
    def contents(self, value):
        if self.check_prop_valid:
            if not isinstance(value, dict) or not all(isinstance(n, awkward.util.string) for n in value):
                raise TypeError("contents must be a dict from strings to arrays")
        for n in list(value):
            value[n] = self._util_toarray(value[n], self.DEFAULTTYPE)
        self._contents = value

    def _getnbytes(self, seen):
        if id(self) in seen:
            return 0
        else:
            seen.add(id(self))
            return sum(x.nbytes if isinstance(x, self.numpy.ndarray) else x._getnbytes(seen) for x in self._contents.values())

    def __len__(self):
        return self._length()

    def _gettype(self, seen):
        out = awkward.type.TableType()
        for n, x in self._contents.items():
            out[n] = awkward.type._fromarray(x, seen)
        return out

    def _util_layout(self, position, seen, lookup):
        args = ()
        for i, (n, x) in enumerate(self._contents.items()):
            awkward.type.LayoutNode(x, position + (i,), seen, lookup)
            args = args + (awkward.type.LayoutArg(n, position + (i,)),)
        return args

    def _length(self):
        if self._view is None:
            if len(self._contents) == 0:
                return 0
            else:
                return min([len(x) for x in self._contents.values()])

        elif isinstance(self._view, tuple):
            start, step, length = self._view
            return length

        else:
            return len(self._view)

    def _index(self):
        if self._view is None:
            return slice(self._length())

        elif isinstance(self._view, tuple):
            start, step, length = self._view
            stop = start + step*length
            if stop < 0:
                stop = None
            return slice(start, stop, step)

        else:
            return self._view

    def _newslice(self, head):
        if self._util_isinteger(head):
            original_head = head

            if self._view is None:
                length = self._length()
                if head < 0:
                    head += length
                if not 0 <= head < length:
                    raise IndexError("index {0} out of bounds for length {1}".format(original_head, length))
                return head

            elif isinstance(self._view, tuple):
                mystart, mystep, mylength = self._view
                if head < 0:
                    head += mylength
                if not 0 <= head < mylength:
                    raise IndexError("index {0} out of bounds for length {1}".format(original_head, mylength))
                return mystart + mystep*head

            else:
                length = len(self._view)
                if head < 0:
                    head += length
                if not 0 <= head < length:
                    raise IndexError("index {0} out of bounds for length {1}".format(original_head, length))
                return self._view[head]

        elif isinstance(head, slice):
            if self._view is None or isinstance(self._view, tuple):
                start, stop, step = head.indices(self._length())
                if step == 0:
                    raise ValueError("slice step cannot be zero")
                if (step > 0 and stop - start > 0) or (step < 0 and stop - start < 0):
                    d, m = divmod(abs(start - stop), abs(step))
                    length = d + (1 if m != 0 else 0)
                else:
                    length = 0

            if self._view is None:
                return start, step, length

            elif isinstance(self._view, tuple):
                mystart, mystep, mylength = self._view
                if step > 0:
                    skip = start
                else:
                    skip = mylength - 1 - start
                return mystart + mystep*start, mystep*step, min(mylength - skip, length)

            else:
                return self._view[head]

        else:
            head = self._util_toarray(head, self.INDEXTYPE)
            if self._util_isintegertype(head.dtype.type):
                length = self._length()
                negative = (head < 0)
                if negative.any():
                    head[negative] += length
                if not self.numpy.bitwise_and(0 <= head, head < length).all():
                    raise IndexError("some indexes out of bounds for length {0}".format(length))

                if self._view is None:
                    return head

                elif isinstance(self._view, tuple):
                    mystart, mystep, mylength = self._view
                    return self.numpy.arange(mystart, mystart + mystep*mylength, mystep)[head]

                else:
                    return self._view[head]

            elif issubclass(head.dtype.type, (self.numpy.bool, self.numpy.bool_)):
                length = self._length()
                if len(head) != length:
                    raise IndexError("boolean index of length {0} does not fit array of length {1}".format(len(head), length))

                if self._view is None:
                    return self.numpy.arange(length)[head]

                elif isinstance(self._view, tuple):
                    mystart, mystep, mylength = self._view
                    return self.numpy.arange(mystart, mystart + mystep*mylength, mystep)[head]

                else:
                    return self._view[head]

            else:
                raise TypeError("cannot interpret dtype {0} as a fancy index or mask".format(head.dtype))

    def _valid(self):
        if self.check_whole_valid:
            pass

    def __iter__(self, checkiter=True):
        if checkiter:
            self._checkiter()

        if self._view is None:
            length = self._length()
            i = 0
            while i < length:
                yield self.Row(self, i)
                i += 1

        elif isinstance(self._view, tuple):
            mystart, mystep, mylength = self._view
            i = 0
            while i < mylength:
                yield self.Row(self, mystart + i*mystep)
                i += 1
        else:
            for i in self._view:
                yield self.Row(self, i)

    def __getitem__(self, where):
        if self._util_isstringslice(where):
            if isinstance(where, awkward.util.string):
                if self._view is None:
                    try:
                        return self._contents[where]
                    except KeyError:
                        raise ValueError("no column named {0}".format(repr(where)))
                else:
                    index = self._index()
                    try:
                        return self._contents[where][index]
                    except KeyError:
                        raise ValueError("no column named {0}".format(repr(where)))
            else:
                contents = OrderedDict()
                for n in where:
                    try:
                        contents[n] = self._contents[n]
                    except KeyError:
                        raise ValueError("no column named {0}".format(repr(n)))
                return self.copy(contents=contents)

        if isinstance(where, tuple) and where == ():
            return self
        if not isinstance(where, tuple):
            where = (where,)
        head, tail = where[0], where[1:]

        if tail != ():
            raise NotImplementedError("multidimensional index through a Table (TODO: needed for [0, n) -> [0, m) -> \"table\" -> ...)")

        newslice = self._newslice(head)

        if self._util_isinteger(newslice):
            return self.Row(self, newslice)

        else:
            out = self.copy(contents=self._contents)
            out._view = newslice
            out._base = self
            out._rowstart = None
            return out

    def __setitem__(self, where, what):
        if self._view is not None:
            raise ValueError("new columns can only be attached to the original Table, not a view (try table.base['col'] = array)")

        if isinstance(where, awkward.util.string):
            try:
                len(what)
            except TypeError:
                what = self.numpy.full(len(self), what)
            self._contents[where] = self._util_toarray(what, self.DEFAULTTYPE)

        elif self._util_isstringslice(where):
            what = what.unzip()
            if len(where) != len(what):
                raise ValueError("number of keys ({0}) does not match number of provided arrays ({1})".format(len(where), len(what)))
            for x, y in zip(where, what):
                try:
                    len(y)
                except TypeError:
                    y = self.numpy.full(len(self), y)
                self._contents[x] = self._util_toarray(y, self.DEFAULTTYPE)

        else:
            raise TypeError("invalid index for assigning column to Table: {0}".format(where))

    def __delitem__(self, where):
        if self._view is not None:
            raise ValueError("columns can only be removed from the original Table, not a view (try del table.base['col'])")

        if isinstance(where, awkward.util.string):
            del self._contents[where]
        elif self._util_isstringslice(where):
            for x in where:
                del self._contents[x]
        else:
            raise TypeError("invalid index for removing column from Table: {0}".format(where))

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if "out" in kwargs:
            raise NotImplementedError("in-place operations not supported")

        if method != "__call__":
            return NotImplemented

        inputsdict = None
        for x in inputs:
            if isinstance(x, Table):
                x._valid()

                if inputsdict is None:
                    inputsdict = OrderedDict([(n, []) for n in x._contents])
                    table = x
                elif set(inputsdict) != set(x._contents):
                    raise ValueError("Tables have different sets of columns")

        assert inputsdict is not None

        for x in inputs:
            if isinstance(x, Table):
                for n, y in x._contents.items():
                    inputsdict[n].append(y[x._index()])
            else:
                for n in inputsdict:
                    inputsdict[n].append(x)

        newcolumns = {}
        tuplelen = None
        for n, x in inputsdict.items():
            newcolumns[n] = getattr(ufunc, method)(*x, **kwargs)

            if tuplelen is None:
                if isinstance(x, tuple):
                    tuplelen = len(x)
                else:
                    tuplelen = False
            elif isinstance(x, tuple) != tuplelen:
                raise AssertionError("ufuncs return tuples of different lengths or some tuples and some non-tuples")

        assert len(newcolumns) != 0
        assert tuplelen is not None

        if self._util_iscomparison(ufunc):
            out = None
            for x in newcolumns.values():
                assert isinstance(x, self.numpy.ndarray)
                assert issubclass(x.dtype.type, (self.numpy.bool_, self.numpy.bool))
                if out is None:
                    out = x
                else:
                    out = self.numpy.bitwise_and(out, x, out=out)
            assert out is not None
            return out

        if method == "at":
            return None

        if tuplelen is False:
            out = table.empty_like()
            for n in inputsdict:
                out[n] = newcolumns[n]
            return out

        else:
            out = [table.empty_like() for i in range(tuplelen)]
            for n in inputsdict:
                for i in range(tuplelen):
                    out[i][n] = newcolumns[n]
            return tuple(out)

    @property
    def counts(self):
        return self.numpy.full(len(self), -1, dtype=self.INDEXTYPE)

    def boolmask(self, maskedwhen=True):
        if maskedwhen:
            return self.numpy.zeros(len(self), dtype=self.MASKTYPE)
        else:
            return self.numpy.ones(len(self), dtype=self.MASKTYPE)

    def choose(self, n):
        raise TypeError("cannot call choose on a Table")

    def argchoose(self, n):
        raise TypeError("cannot call argchoose on a Table")

    def distincts(self, nested=False):
        raise TypeError("cannot call distincts on a Table")

    def argdistincts(self, nested=False):
        raise TypeError("cannot call argdistincts on a Table")

    def pairs(self, nested=False):
        raise TypeError("cannot call pairs on a Table")

    def argpairs(self, nested=False):
        raise TypeError("cannot call argpairs on a Table")

    def cross(self, other, nested=False):
        raise TypeError("cannot call cross on a Table")

    def argcross(self, other, nested=False):
        raise TypeError("cannot call argcross on a Table")

    def flatten(self, axis=0):
        raise ValueError("cannot flatten through a Table")

    def pad(self, length, maskedwhen=True, clip=False, axis=0):
        out = self.copy(contents={})
        for n, x in self._contents.items():
            out[n] = self._util_pad(x, length, maskedwhen, clip, axis)
        return out

    def regular(self):
        self._valid()
        pairs = [(n, self._util_regular(x)) for n, x in self.items()]
        out = self.numpy.empty(len(self), [(n, x.dtype, x.shape[1:]) for n, x in pairs])
        for n, x in pairs:
            out[n] = x
        return out

    def flattentuple(self):
        out = self.copy()
        out._contents = OrderedDict([(n, x.flattentuple() if isinstance(x, Table) else x) for n, x in out._contents.items()])

        if self.istuple:
            contents = OrderedDict()
            for n, x in out._contents.items():
                if isinstance(x, Table) and x.istuple:
                    if x._view is None:
                        view = slice(x._length())

                    elif isinstance(x._view, tuple):
                        start, step, length = x._view
                        stop = start + step*length
                        if stop < 0:
                            stop = None
                        view = slice(start, stop, step)

                    else:
                        view = x._view

                    for y in x._contents.values():
                        contents[str(len(contents))] = y[view]

                else:
                    contents[str(len(contents))] = x

            out._contents = contents

        return out

    def _hasjagged(self):
        num = sum(1 if self._util_hasjagged(x) else 0 for x in self._contents.values())
        if num == 0:
            return False
        elif num == len(self._contents):
            return True
        else:
            raise ValueError("some Table columns are jagged and others are not")

    def _reduce(self, ufunc, identity, dtype):
        if self._hasjagged():
            out = self.copy(contents={})
            for n, x in self._contents.items():
                out[n] = self._util_reduce(x, ufunc, identity, dtype)
            return out

        else:
            out = self.Table.named({
                self.numpy.bitwise_or: "any",
                self.numpy.bitwise_and: "all",
                None: "count",
                self.numpy.count_nonzero: "count_nonzero",
                self.numpy.add: "sum",
                self.numpy.multiply: "prod",
                self.numpy.minimum: "min",
                self.numpy.maximum: "max"
                }[ufunc])
            out._showdict = True
            for n in self._contents:
                x = self._contents[n][self._index()]
                out[n] = self.numpy.array([self._util_reduce(x, ufunc, identity, dtype)])
            return out.Row(out, 0)

    def _prepare(self, ufunc, identity, dtype):
        out = self.copy(contents={})
        for n, x in self._contents.items():
            if isinstance(x, self.numpy.ndarray):
                if dtype is None and issubclass(x.dtype.type, (self.numpy.bool_, self.numpy.bool)):
                    dtype = self.numpy.dtype(type(identity))
                if dtype is not None:
                    x = x.astype(dtype)
            else:
                x = x._prepare(ufunc, identity, dtype)
            out[n] = x
        return out

    def argmin(self):
        raise TypeError("cannot call argmin on Table")

    def argmax(self):
        raise TypeError("cannot call argmax on Table")

    def _util_columns(self, seen):
        if id(self) in seen:
            return []
        seen.add(id(self))
        return list(self._contents)

    def astype(self, dtype):
        out = self.copy(contents={})
        for n, x in self._contents.items():
            out[n] = x.astype(dtype)
        return out

    def fillna(self, value):
        out = self.copy(contents={})
        for n, x in self._contents.items():
            out[n] = self._util_fillna(x, value)
        return out

    @classmethod
    def _concatenate_axis0(cls, tables):
        for i in range(len(tables)-1):
            if set(tables[i]._contents) != set(tables[i+1]._contents):
                raise ValueError("cannot concatenate Tables with different fields")

        out = tables[0].deepcopy(contents=OrderedDict())

        for n in tables[0]._contents:
            content_type = type(tables[0]._contents[n])
            if content_type == cls.numpy.ndarray:
                concatenate = cls.numpy.concatenate
            else:
                concatenate = content_type.concatenate

            out._contents[n] = concatenate([t[n] for t in tables], axis=0)

        out._valid()
        return out

    _topandas_name = "TableSeries"

    def _topandas(self, seen):
        import awkward.pandas
        if id(self) in seen:
            return seen[id(self)]
        else:
            out = seen[id(self)] = self.copy()
            out.__class__ = awkward.pandas.mixin(type(self))
            out._contents = OrderedDict((n, x._topandas(seen) if isinstance(x, awkward.array.base.AwkwardArray) else x) for n, x in out._contents.items())
            return out



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import awkward.array.union
from .base import NumbaMethods

class UnionArrayNumba(NumbaMethods, awkward.array.union.UnionArray):
    pass



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import os.path

from setuptools import find_packages
from setuptools import setup

def get_version():
    g = {}
    exec(open(os.path.join("..", "awkward", "version.py")).read(), g)
    return g["__version__"]

setup(name = "awkward-numba",
      version = get_version(),
      packages = find_packages(exclude = ["tests"]),
      scripts = [],
      description = "Allows awkward arrays to be used in Numba-compiled code and optimizes awkward methods with JIT compilation.",
      long_description = """.. image:: https://raw.githubusercontent.com/scikit-hep/awkward-array/master/docs/source/logo-300px.png
   :alt: awkward-array
   :target: https://github.com/scikit-hep/awkward-array

|

Allows awkward arrays to be used in Numba-compiled code and optimizes awkward methods with JIT compilation.

See tests/study_numba_speed.py for an example.

Note: be sure to ``import awkward.numba`` before attempting to use an awkward array in a Numba routine.

Status: only JaggedArrays have been implemented.
""",
      author = "Jim Pivarski (IRIS-HEP)",
      author_email = "pivarski@princeton.edu",
      maintainer = "Jim Pivarski (IRIS-HEP)",
      maintainer_email = "pivarski@princeton.edu",
      url = "https://github.com/scikit-hep/awkward-array",
      download_url = "https://github.com/scikit-hep/awkward-array/releases",
      license = "BSD 3-clause",
      test_suite = "tests",
      install_requires = ["awkward==" + get_version(), "numba"],
      setup_requires = ["pytest-runner"],
      tests_require = ["pytest"],
      classifiers = [
          # "Development Status :: 1 - Planning",
          # "Development Status :: 2 - Pre-Alpha",
          "Development Status :: 3 - Alpha",
          # "Development Status :: 4 - Beta",
          # "Development Status :: 5 - Production/Stable",
          # "Development Status :: 6 - Mature",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ],
      platforms = "Any",
      )



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import base64
import fnmatch
import importlib
import json
import numbers
import os
import pickle
import types
import zipfile
import zlib
from itertools import count
try:
    from collections.abc import Mapping, MutableMapping
except ImportError:
    from collections import Mapping, MutableMapping

import numpy

import awkward.type
import awkward.version

compression = [
        {"minsize": 8192, "types": [numpy.bool_, numpy.bool, numpy.integer], "contexts": "*", "pair": (zlib.compress, ("zlib", "decompress"))},
    ]

whitelist = [
        ["numpy", "frombuffer"],
        ["zlib", "decompress"],
        ["lzma", "decompress"],
        ["backports.lzma", "decompress"],
        ["lz4.block", "decompress"],
        ["awkward", "*Array"],
        ["awkward", "Table"],
        ["awkward", "numpy", "frombuffer"],
        ["awkward.util", "frombuffer"],
        ["awkward.persist"],
        ["awkward.arrow", "_ParquetFile", "fromjson"],
        ["uproot_methods.classes.*"],
        ["uproot_methods.profiles.*"],
        ["uproot.tree", "_LazyFiles"],
        ["uproot.tree", "_LazyTree"],
        ["uproot.tree", "_LazyBranch"],
    ]

def frompython(obj):
    return base64.b64encode(pickle.dumps(obj)).decode("ascii")

def topython(string):
    return pickle.loads(base64.b64decode(string.encode("ascii")))

def spec2function(obj, awkwardlib="awkward", whitelist=whitelist):
    for white in whitelist:
        for n, p in zip(obj, white):
            if not fnmatch.fnmatchcase(n, p):
                break
        else:
            if obj[0] == "awkward":
                obj = [awkwardlib] + obj[1:]
            gen, genname = importlib.import_module(obj[0]), obj[1:]
            if not isinstance(gen, types.ModuleType):
                raise TypeError("first item of a function description must be a module")
            if genname[:1] == ["numpy"]:
                gen, genname = getattr(gen, genname[0]), genname[1:]
            while len(genname) > 0:
                gen, genname = getattr(gen, genname[0]), genname[1:]
                if isinstance(gen, types.ModuleType):
                    raise TypeError("non-first items of a function description must not be a module")
            break
    else:
        raise RuntimeError("callable not in whitelist; add it by passing a whitelist argument:\n\n    whitelist = awkward.persist.whitelist + [{0}]".format(repr(obj)))
    return gen

def dtype2json(obj):
    if obj.subdtype is not None:
        dt, sh = obj.subdtype
        return (dtype2json(dt), sh)
    elif obj.names is not None:
        return [(n, dtype2json(obj[n])) for n in obj.names]
    else:
        return str(obj)

def json2dtype(obj):
    def recurse(obj):
        if isinstance(obj, (list, tuple)) and len(obj) > 0 and (isinstance(obj[-1], numbers.Integral) or isinstance(obj[0], str) or (isinstance(obj[-1], (list, tuple)) and all(isinstance(x, numbers.Integral) for x in obj[-1]))):
            return tuple(recurse(x) for x in obj)
        elif isinstance(obj, (list, tuple)):
            return [recurse(x) for x in obj]
        else:
            return obj
    return numpy.dtype(recurse(obj))

def type2json(obj):
    if isinstance(obj, awkward.type.Type):
        labeled = obj._labeled()
    else:
        labeled = []

    seen = set()

    def takes(n):
        if n == float("inf"):
            return "inf"
        else:
            return int(n)

    def recurse(obj):
        if isinstance(obj, awkward.type.Type):
            if id(obj) in seen:
                for i, x in enumerate(labeled):
                    if obj is x:
                        return {"ref": "T{0}".format(i)}

            else:
                seen.add(id(obj))
                if isinstance(obj, awkward.type.ArrayType):
                    out = {"takes": takes(obj._takes), "to": recurse(obj._to)}

                elif isinstance(obj, awkward.type.TableType):
                    out = {"fields": [[n, recurse(x)] for n, x in obj._fields.items()]}

                elif isinstance(obj, awkward.type.UnionType):
                    out = {"possibilities": [recurse(x) for x in obj._possibilities]}

                elif isinstance(obj, awkward.type.OptionType):
                    out = {"type": recurse(obj._type)}

                for i, x in enumerate(labeled):
                    if obj is x:
                        return {"set": "T{0}".format(i), "as": out}
                else:
                    return out

        elif isinstance(obj, numpy.dtype):
            return {"dtype": dtype2json(obj)}

        elif callable(obj):
            if obj.__module__ == "__main__":
                raise TypeError("cannot persist object type: its generator is defined in __main__, which won't be available in a subsequent session")
            if hasattr(obj, "__qualname__"):
                spec = [obj.__module__] + obj.__qualname__.split(".")
            else:
                spec = [obj.__module__, obj.__name__]

            gen, genname = importlib.import_module(spec[0]), spec[1:]
            while len(genname) > 0:
                gen, genname = getattr(gen, genname[0]), genname[1:]
            if gen is not obj:
                raise TypeError("cannot persist object type: its generator cannot be found via its __name__ (Python 2) or __qualname__ (Python 3)")

            return {"function": spec}

        else:
            raise TypeError("only awkward.type.Type, numpy.dtype, and callables are types")

    return recurse(obj)

def json2type(obj, whitelist=whitelist):
    labels = {}

    def takes(n):
        if n == "inf":
            return float("inf")
        else:
            return n

    def recurse(obj):
        if not isinstance(obj, dict):
            raise TypeError("json2type is expecting a JSON object, found: {0}".format(repr(obj)))

        if "set" in obj:
            placeholder = labels[obj["set"]] = awkward.type.Placeholder()
            placeholder.value = recurse(obj["as"])
            return placeholder

        elif "ref" in obj:
            return labels[obj["ref"]]

        elif "takes" in obj and "to" in obj:
            return awkward.type.ArrayType(takes(obj["takes"]), recurse(obj["to"]))

        elif "fields" in obj:
            out = awkward.type.TableType()
            for n, x in obj["fields"]:
                out[n] = recurse(x)
            return out

        elif "possibilities" in obj:
            return awkward.type.UnionType(*[recurse(x) for x in obj["possibilities"]])

        elif "type" in obj:
            return awkward.type.OptionType(recurse(obj["type"]))

        elif "dtype" in obj:
            return json2dtype(obj["dtype"])

        elif "function" in obj:
            return spec2function(obj["function"], whitelist=whitelist)

        else:
            raise ValueError("unexpected set of keys in JSON: {0}".format(", ".join(repr(x) for x in obj)))

    return awkward.type._resolve(recurse(obj), {})

def jsonable(obj):
    if obj is None:
        return obj

    elif isinstance(obj, dict) and all(isinstance(n, str) for n in obj):
        return {n: jsonable(x) for n, x in obj.items()}

    elif isinstance(obj, list):
        return [jsonable(x) for x in obj]

    elif isinstance(obj, str):
        return str(obj)

    elif isinstance(obj, (bool, numpy.bool_, numpy.bool)):
        return bool(obj)      # policy: eliminate Numpy types

    elif isinstance(obj, (numbers.Integral, numpy.integer)):
        return int(obj)       # policy: eliminate Numpy types

    elif isinstance(obj, (numbers.Real, numpy.floating)) and numpy.isfinite(obj):
        return float(obj)     # policy: eliminate Numpy types

    else:
        raise TypeError("object cannot be losslessly serialized as JSON")

class ObjRef(object):
    def __init__(self, idgen=None):
        if idgen:
            self.idgen = iter(idgen)
        self._i2r = {}
        self._r2o = {}

    def nextid(self):
        return next(self.idgen)

    def __contains__(self, obj):
        return id(obj) in self._i2r

    def __setitem__(self, obj, ref):
        self._i2r[id(obj)] = ref
        self._r2o[ref] = obj

    def __getitem__(self, obj):
        if obj not in self:
            self[obj] = self.nextid()
        return self._i2r[id(obj)]

    def __delitem__(self, obj):
        assert obj in self
        del self._r2o[self._i2r[id(obj)]]
        del self._i2r[id(obj)]

    def get(self, obj, default=None):
        return self[obj] if obj in self else default

    def obj(self, ref):
        if ref in self._r2o:
            return self._r2o[ref]
        else:
            return awkward.array.virtual.VirtualArray(lambda: self._r2o[ref])

class Serializer(object):
    def __init__(self, storage, prefix="", suffix="", schemasuffix=""):
        self.storage = storage
        self.suffix = suffix
        self.prefix = prefix
        self.schemasuffix = schemasuffix
        self.seen = ObjRef(idgen=count())

    def store(self, name, obj):
        schema = {"awkward": awkward.version.__version__, "schema": self(obj)}
        if self.prefix != "":
            schema["prefix"] = self.prefix

        schema = self._finalize_schema(schema) or schema

        self.storage[name + self.schemasuffix] = self._encode_schema(schema)
        return schema

    def load(self, *args, **kwargs):
        return deserialize(*args, storage=self.storage, seen=self.seen, **kwargs)

    def encode_call(self, *args, **kwargs):
        func, args = args[0], args[1:]
        out = {"call": self._obj2spec(func) if callable(func) else tuple(func)}
        if args:
            out["args"] = list(args)
        if kwargs:
            out["kwargs"] = kwargs
        return out

    def encode_json(self, obj):
        return {"json": jsonable(obj)}

    def encode_python(self, obj):
        return {"python": frompython(obj)}

    @classmethod
    def _encode_primitive(cls, obj):
        if isinstance(obj, numpy.dtype):
            return {"dtype": dtype2json(obj)}

    @classmethod
    def _obj2spec(cls, obj, test=True):
        if hasattr(obj, "__qualname__"):
            spec = [obj.__module__] + obj.__qualname__.split(".")
        else:
            spec = [obj.__module__, obj.__name__]

        if test:
            val = importlib.import_module(spec[0])
            for key in spec[1:]:
                val = getattr(val, key)
            assert val == obj

        return spec

    def _encode_complex(self, obj, context):
        if callable(getattr(obj, "__awkward_serialize__", None)):
            return obj.__awkward_serialize__(self)

        if hasattr(obj, "tojson") and hasattr(type(obj), "fromjson"):
            try:
                return self.encode_call(self._obj2spec(type(obj).fromjson), self.encode_json(obj.tojson()))
            except:
                pass

        if isinstance(obj, numpy.ndarray):
            return self._encode_numpy(obj, context)

        if hasattr(obj, "__module__") and (hasattr(obj, "__qualname__") or hasattr(obj, "__name__")) and obj.__module__ != "__main__":
            try:
                return {"function": self._obj2spec(obj)}
            except:
                pass

        try:
            return self.encode_json(obj)
        except TypeError:
            pass

        try:
            return self.encode_python(obj)
        except:
            pass

    def _encode_numpy(self, obj, context):
        key = str(self.seen[obj]) + self.suffix
        self.storage[self.prefix + key] = obj
        return {"read": key}

    def _encode_schema(self, schema):
        return json.dumps(schema).encode("ascii")

    def _finalize_schema(self, schema):
        pass

    def __call__(self, obj, context=""):
        out = self._encode_primitive(obj)

        if out is not None:
            return out

        if obj in self.seen:
            return {"ref": self.seen[obj]}
        else:
            ident = self.seen[obj]

        out = self._encode_complex(obj, context)
        if out is None:
            raise TypeError("failed to encode {0} (type: {1})".format(repr(obj), type(obj)))

        if "id" in out:
            if out["id"] is False:
                del self.seen[obj]
            elif out["id"] != self.seen[obj]:
                raise RuntimeError("unexpected id change")
        else:
            out["id"] = ident

        return out

    def fill(self, obj, context, prefix, suffix, schemasuffix, storage, compression, **kwargs):
        assert self.prefix == prefix
        assert self.suffix == suffix
        assert self.schemasuffix == schemasuffix
        assert self.storage == storage
        assert self.compression == compression
        return self(obj, context=context)

class BlobSerializer(Serializer):
    class CompressPolicy(object):
        enc2dec = {
            zlib.compress: ("zlib", "decompress"),
        }

        @classmethod
        def parse(cls, x):
            if isinstance(x, cls):
                return x
            elif isinstance(x, dict):
                return cls(**x)
            elif callable(x):
                return cls(enc=x)
            elif len(x) == 2 and callable(x[0]):
                return cls(enc=x[0], dec=x[1])
            else:
                raise TypeError("can't parse compression policy {0}".format(x))

        def __init__(self, pair=None, enc=None, dec=None, minsize=0, types=object, contexts="*"):
            if pair is not None:
                enc, dec = pair
            if dec is None:
                dec = self.enc2dec[enc]
            if isinstance(types, list):
                types = tuple(types)
            elif not isinstance(types, tuple):
                types = types,
            if isinstance(contexts, list):
                contexts = tuple(contexts)
            elif not isinstance(contexts, tuple):
                contexts = contexts,
            assert callable(enc)
            assert isinstance(dec, tuple)
            assert 0 <= minsize
            self.enc = enc
            self.dec = dec
            self.minsize = minsize
            self.types = types
            self.contexts = contexts

        @property
        def pair(self):
            return (self.enc, self.dec)

        def test(self, obj, context):
            return (obj.nbytes >= self.minsize and issubclass(obj.dtype.type, tuple(self.types)) and any(fnmatch.fnmatchcase(context, p) for p in self.contexts))

    @classmethod
    def _parse_compression(cls, comp):
        if comp is None or comp is False:
            comp = []
        elif comp is True:
            comp = [{"minsize": 0, "types": object, "contexts": "*", "pair": (zlib.compress, ("zlib", "decompress"))}]
        elif not isinstance(comp, (list, tuple)):
            comp = [comp]

        return list(map(cls.CompressPolicy.parse, comp))

    def __init__(self, *args, **kwargs):
        self.compression = self._parse_compression(kwargs.pop("compression", compression))
        super(BlobSerializer, self).__init__(*args, **kwargs)

    def _put_raw(self, data, ref=None):
        if ref is None:
            ref = data
        key = str(self.seen[ref]) + self.suffix
        self.storage[self.prefix + key] = data
        return dict(read=key)

    def _encode_numpy(self, obj, context):
        if obj.ndim > 1:
            dtype = numpy.dtype((obj.dtype, obj.shape[1:]))
        else:
            dtype = obj.dtype

        buf = None
        for policy in self._parse_compression(self.compression):
            if policy.test(obj, context):
                buf = self.encode_call(policy.dec, self._put_raw(policy.enc(obj.ravel()), ref=obj))
                break
        else:
            buf = self._put_raw(obj.ravel(), ref=obj)

        return self.encode_call(["awkward", "numpy", "frombuffer"], buf, self(dtype), self(obj.shape[0]))

def serialize(obj, storage, name="", delimiter="-", **kwargs):
    if delimiter is None:
        delimiter = ""
    if name:
        kwargs.setdefault("prefix", name + delimiter)
    return BlobSerializer(storage, **kwargs).store(name, obj)

def deserialize(storage, name="", awkwardlib="awkward", whitelist=whitelist, cache=None, seen=None):
    import awkward.array.virtual

    schema = storage[name]
    if isinstance(schema, numpy.ndarray):
        schema = schema.tostring()
    if isinstance(schema, bytes):
        schema = schema.decode("ascii")
    schema = json.loads(schema)

    if "awkward" not in schema:
        raise ValueError("JSON object is not an awkward-array schema (missing 'awkward' field)")

    prefix = schema.get("prefix", "")
    if seen is None:
        seen = ObjRef()

    if isinstance(whitelist, str):
        whitelist = [whitelist]
    elif len(whitelist) > 0 and isinstance(whitelist[0], str):
        whitelist = [whitelist]

    def unfill(schema):
        if isinstance(schema, dict):
            if "call" in schema and isinstance(schema["call"], list) and len(schema["call"]) > 0:
                gen = spec2function(schema["call"], awkwardlib=awkwardlib, whitelist=whitelist)
                args = [unfill(x) for x in schema.get("args", [])]

                kwargs = {}
                if schema.get("cacheable", False):
                    kwargs["cache"] = cache
                if schema.get("whitelistable", False):
                    kwargs["whitelist"] = whitelist
                if "kwargs" in schema:
                    kwargs.update({n: unfill(x) for n, x in schema["kwargs"].items()})

                out = gen(*args, **kwargs)

            elif "read" in schema:
                if schema.get("absolute", False):
                    out = storage[schema["read"]]
                else:
                    out = storage[prefix + schema["read"]]

            elif "list" in schema:
                out = [unfill(x) for x in schema["list"]]

            elif "tuple" in schema:
                out = tuple(unfill(x) for x in schema["tuple"])

            elif "dict" in schema:
                out = {n: unfill(x) for n, x in schema["dict"].items()}

            elif "pairs" in schema:
                out = [(n, unfill(x)) for n, x in schema["pairs"]]

            elif "dtype" in schema:
                out = json2dtype(schema["dtype"])

            elif "function" in schema:
                out = spec2function(schema["function"], awkwardlib=awkwardlib, whitelist=whitelist)

            elif "json" in schema:
                out = schema["json"]

            elif "python" in schema:
                out = topython(schema["python"])

            elif "ref" in schema:
                out = seen.obj(schema["ref"])

            else:
                raise ValueError("unrecognized JSON object with fields {0}".format(", ".join(repr(x) for x in schema)))

            if "id" in schema:
                seen[out] = schema["id"]
            return out

        elif isinstance(schema, list):
            raise ValueError("unrecognized JSON list with length {0}".format(len(schema)))

        else:
            raise ValueError("unrecognized JSON object: {0}".format(repr(schema)))

    return unfill(schema["schema"])

def keys(storage, name="", subschemas=True):
    schema = storage[name]
    if isinstance(schema, numpy.ndarray):
        schema = schema.tostring()
    if isinstance(schema, bytes):
        schema = schema.decode("ascii")
    schema = json.loads(schema)

    prefix = schema.get("prefix", "")

    def recurse(schema):
        if isinstance(schema, dict):
            if "call" in schema and isinstance(schema["call"], list) and len(schema["call"]) > 0:
                for x in schema.get("args", []):
                    for y in recurse(x):
                        yield y
                for x in schema.get("kwargs", {}).values():
                    for y in recurse(x):
                        yield y
                for x in schema.get("*", []):
                    for y in recurse(x):
                        yield y
                for x in schema.get("**", {}).values():
                    for y in recurse(x):
                        yield y

            elif "read" in schema:
                if schema.get("absolute", False):
                    yield schema["read"]
                else:
                    yield prefix + schema["read"]

            elif "list" in schema:
                for x in schema["list"]:
                    for y in recurse(x):
                        yield y

            elif "tuple" in schema:
                for x in schema["tuple"]:
                    for y in recurse(x):
                        yield y

            elif "dict" in schema:
                for x in schema["dict"].values():
                    for y in recurse(x):
                        yield y

            elif "pairs" in schema:
                for n, x in schema["pairs"]:
                    for y in recurse(x):
                        yield y

            elif "dtype" in schema:
                pass

            elif "function" in schema:
                pass

            elif "json" in schema:
                pass

            elif "python" in schema:
                pass

            elif "ref" in schema:
                pass

    yield name
    for x in recurse(schema["schema"]):
        yield x

def save(file, array, name=None, mode="a", **options):
    if isinstance(array, dict):
        arrays = array
    else:
        arrays = {"": array}

    if name is not None:
        arrays = {name + n: x for n, x in arrays.items()}

    arraynames = list(arrays)
    for i in range(len(arraynames)):
        for j in range(i + 1, len(arraynames)):
            if arraynames[i].startswith(arraynames[j]) or arraynames[j].startswith(arraynames[i]):
                raise KeyError("cannot write both {0} and {1} to zipfile because one is a prefix of the other", repr(arraynames[i]), repr(arraynames[j]))

    if isinstance(file, getattr(os, "PathLike", ())):
        file = os.fspath(file)
    elif hasattr(file, "__fspath__"):
        file = file.__fspath__()
    elif file.__class__.__module__ == "pathlib":
        import pathlib
        if isinstance(file, pathlib.Path):
             file = str(file)

    if isinstance(file, str) and not file.endswith(".awkd"):
        file = file + ".awkd"

    alloptions = {"delimiter": "-", "suffix": ".raw", "schemasuffix": ".json", "compression": compression}
    alloptions.update(options)
    options = alloptions

    class Wrap(object):
        def __init__(self, f):
            self.f = f
        def __setitem__(self, where, what):
            if isinstance(what, numpy.ndarray):
                what = what.tostring()
            self.f.writestr(where, what, compress_type=zipfile.ZIP_STORED)

    with zipfile.ZipFile(file, mode=mode, compression=zipfile.ZIP_STORED) as f:
        namelist = f.namelist()
        for name in arraynames:
            if any(n.startswith(name) for n in namelist):
                raise KeyError("cannot add {0} to zipfile because the following already exist: {1}".format(repr(name), ", ".join(repr(n) for n in namelist if n.startswith(name))))

        wrapped = Wrap(f)
        for name, array in arrays.items():
            serialize(array, wrapped, name=name, **options)

def load(file, **options):
    f = Load(file, **options)
    if list(f) == [""]:
        out = f[""]
        f.close()
        return out
    else:
        return f

class Load(Mapping):
    def __init__(self, file, **options):
        class Wrap(object):
            def __init__(self):
                self.f = zipfile.ZipFile(file, mode="r")
            def __getitem__(self, where):
                return self.f.read(where)

        self._file = Wrap()

        alloptions = {"schemasuffix": ".json", "awkwardlib": "awkward", "whitelist": whitelist, "cache": None}
        alloptions.update(options)
        self.schemasuffix = alloptions.pop("schemasuffix")
        self.options = alloptions

    def __getitem__(self, where):
        return deserialize(self._file, name=where + self.schemasuffix, awkwardlib=self.options["awkwardlib"], whitelist=self.options["whitelist"], cache=self.options["cache"])

    def __iter__(self):
        for n in self._file.f.namelist():
            if n.endswith(".json"):
                yield n[:-5]

    def __len__(self):
        count = 0
        for n in self._file.f.namelist():
            if n.endswith(".json"):
                count += 1
        return count

    def __repr__(self):
        return "<awkward.load ({0} members)>".format(len(self))

    def close(self):
        self._file.f.close()

    def __del__(self):
        self.close()

    def __enter__(self, *args, **kwds):
        return self

    def __exit__(self, *args, **kwds):
        self.close()

class hdf5(MutableMapping):
    def __init__(self, group, **options):
        alloptions = {"compression": compression, "awkwardlib": "awkward", "whitelist": whitelist, "cache": None}
        alloptions.update(options)
        self.options = alloptions
        self.options["delimiter"] = "/"
        self.options["schemasuffix"] = "/schema.json"

        class Wrap(object):
            def __init__(self):
                self.g = group
            def __getitem__(self, where):
                return self.g[where][()]
            def __setitem__(self, where, what):
                self.g[where] = numpy.frombuffer(what, dtype=numpy.uint8)

        self._group = Wrap()

    def __getitem__(self, where):
        return deserialize(self._group, name=where + self.options["schemasuffix"], awkwardlib=self.options["awkwardlib"], whitelist=self.options["whitelist"], cache=self.options["cache"])

    def __setitem__(self, where, what):
        options = dict(self.options)
        if "awkwardlib" in options:
            del options["awkwardlib"]
        if "whitelist" in options:
            del options["whitelist"]
        if "cache" in options:
            del options["cache"]
        self._group.g.create_group(where)
        serialize(what, self._group, name=where, **options)

    def __delitem__(self, where):
        for subname in keys(self._group, name=where + self.options["schemasuffix"]):
            del self._group.g[subname]
        del self._group.g[where]

    def __iter__(self):
        schemaname = self.options["schemasuffix"].split("/")[-1]
        for subname in self._group.g:
            if schemaname in self._group.g[subname]:
                yield subname

    def __len__(self):
        schemaname = self.options["schemasuffix"].split("/")[-1]
        count = 0
        for subname in self._group.g:
            if schemaname in self._group.g[subname]:
                count += 1
        return count

    def __repr__(self):
        return "<awkward.hdf5 {0} ({1} members)>".format(repr(self._group.g.name), len(self))



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import awkward.array.virtual
from .base import NumbaMethods

class VirtualArrayNumba(NumbaMethods, awkward.array.virtual.VirtualArray):
    pass



import operator

import numpy
import numba

class Dependent(object):
    def __init__(self, **available):
        self.available = available

    def __getitem__(self, where):
        return self.available[where]

@numba.extending.typeof_impl.register(Dependent)
def _Dependent_typeof(val, c):
    return DependentType(list(val.available))

class DependentType(numba.types.Type):
    def __init__(self, available):
        self.available = available
        self.requested = set()
        super(DependentType, self).__init__(name="DependentType({0})".format(", ".join(sorted(self.available))))

    def request(self, name):
        if name not in self.requested:
            self.requested.add(name)

@numba.typing.templates.infer
class _DependentType_type_getitem(numba.typing.templates.AbstractTemplate):
    key = "static_getitem"
    def generic(self, args, kwargs):
        if len(args) == 2 and len(kwargs) == 0:
            objtype, where = args
            print("checking type", objtype, where)
            objtype.request(where)
            return numba.types.int64

@numba.extending.register_model(DependentType)
class DependentModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        print("making model", fe_type)
        print("requested", fe_type.requested)
        members = []
        super(DependentModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(DependentType)
def _JaggedArray_unbox(typ, obj, c):
    print("unboxing", typ)
    print("requested", typ.requested)
    out = numba.cgutils.create_struct_proxy(typ)(c.context, c.builder)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(out._getvalue(), is_error)

@numba.extending.lower_builtin("static_getitem", DependentType, numba.types.StringLiteral)
def _DependentType_lower_static_getitem(context, builder, sig, args):
    print("lowering", sig.args[0], sig.args[1].literal_value)
    return context.get_constant(numba.types.int64, 999)

obj = Dependent(one=999, two=999, three=999)

@numba.njit
def f(x):
    return x["one"]
print(f(obj))
print(f(obj))

@numba.njit
def g(x):
    return x["two"]
print(g(obj))
print(g(obj))
print(f(obj))

obj = Dependent(one=999, two=999, three=999)

print(f(obj))
print(g(obj))

obj = Dependent(one=999, two=999, three=999)

@numba.njit
def h(x):
    return f(x)
print(h(obj))



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import re
from collections import OrderedDict

import numpy
import numba

import awkward.array.table
from .base import NumbaMethods
from .base import AwkwardArrayType
from .base import clsrepr
from .base import sliceval2
from .base import sliceval3

class TableNumba(NumbaMethods, awkward.array.table.Table):
    class Row(awkward.array.table.Table.Row):
        pass

        # def __init__(self, table, index):

        # def __repr__(self):

        # def __contains__(self, name):

        # def tolist(self):

        # def __len__(self):

        # def __iter__(self, checkiter=True):

        # def __getitem__(self, where):

        # def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

        # def __eq__(self, other):

        # def __ne__(self, other):

        # @property
        # def i0(self):

        # @property
        # def i1(self):

        # @property
        # def i2(self):

        # @property
        # def i3(self):

        # @property
        # def i4(self):

        # @property
        # def i5(self):

        # @property
        # def i6(self):

        # @property
        # def i7(self):

        # @property
        # def i8(self):

        # @property
        # def i9(self):

    # def __init__(self, columns1={}, *columns2, **columns3):

    # def tolist(self):

    # @classmethod
    # def named(cls, rowname, columns1={}, *columns2, **columns3):

    # @property
    # def rowname(self):

    # @rowname.setter
    # def rowname(self, value):

    # @classmethod
    # def fromrec(cls, recarray):

    # @classmethod
    # def frompairs(cls, pairs):

    # @classmethod
    # def fromview(cls, view, base):

    # def copy(self, contents=None):

    # def deepcopy(self, contents=None):

    # def empty_like(self, **overrides):

    # def zeros_like(self, **overrides):

    # def ones_like(self, **overrides):

    # def __awkward_persist__(self, ident, fill, prefix, suffix, schemasuffix, storage, compression, **kwargs):

    # @property
    # def base(self):

    # @property
    # def contents(self):

    # @contents.setter
    # def contents(self, value):

    # def __len__(self):

    # def _gettype(self, seen):

    # def _length(self):

    # def _index(self):

    # def _newslice(self, head):

    # def _valid(self):

    # def __iter__(self, checkiter=True):

    # def __getitem__(self, where):

    # def __setitem__(self, where, what):

    # def __delitem__(self, where):

    # def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

    # @property
    # def istuple(self):

    # def flattentuple(self):

    # def _hasjagged(self):

    # def _reduce(self, ufunc, identity, dtype, regularaxis):

    # @property
    # def columns(self):

    # def astype(self, dtype):

######################################################################## register types in Numba

@numba.extending.typeof_impl.register(awkward.array.table.Table)
def _Table_typeof(val, c):
    return TableType(val.rowname, OrderedDict((n, numba.typeof(x)) for n, x in val.contents.items()), special=type(val), specialrow=type(val).Row)

class TableType(AwkwardArrayType):
    class RowType(numba.types.Type):
        def __init__(self, rowname, types, special=awkward.array.table.Table.Row):
            super(RowType, self).__init__(name="RowType({0}, {{{1}}}{2})".format(repr(rowname), ", ".join("{0}: {1}".format(repr(n), t) for n, t in types.items()), "" if special is awkward.array.table.Table.Row else clsrepr(special)))
            self.rowname = rowname
            self.types = types
            self.special = special

    @property
    def rowtype(self):
        return TableType.RowType(self.rowname, self.types, special=self.specialrow)

    def __init__(self, rowname, types, special=awkward.array.table.Table, specialrow=awkward.array.table.Table.Row):
        super(TableType, self).__init__(name="TableType({0}, {{{1}}}{2}{3})".format(repr(rowname), ", ".join("{0}: {1}".format(repr(n), t) for n, t in types.items()), "" if special is awkward.array.table.Table else clsrepr(special), "" if specialrow is awkward.array.table.Table.Row else clsrepr(specialrow)))
        self.rowname = rowname
        self.types = types
        self.special = special
        self.specialrow = specialrow

    def getitem(self, wheretype):
        if isinstance(wheretype, numba.types.StringLiteral):
            if wheretype.literal_value in self.types:
                return self.types[wheretype.literal_value]
            else:
                raise TypeError("{0} is not a column of Table".format(repr(wheretype.literal_value)))

        # elif literal list/tuple of strings

        else:
            return TableType(self.rowname, OrderedDict((n, x.getitem(wheretype)) for n, x in self.types.items()), special=self.special, specialrow=self.specialrow)

######################################################################## model and boxing

# don't start non-fieldnames with "f"
def _safename(name):
    return "f" + _safename._pattern.sub(lambda bad: "_" + "".join("{0:02x}".format(ord(x)) for x in bad.group(0)) + "_", name)
_safename._pattern = re.compile("[^a-zA-Z0-9]+")

@numba.extending.register_model(TableType)
class TableModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [(_safename(n), x) for n, x in fe_type.types.items()]
        super(TableModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(TableType)
def _TableType_unbox(typ, obj, c):
    key_objs = [HERE]
    value_objs = [c.pyapi.object_getitem(obj, n) for n in key_objs]



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import numpy

from awkward.numba.array.base import NumbaMethods
from awkward.numba.array.chunked import ChunkedArrayNumba as ChunkedArray
from awkward.numba.array.chunked import AppendableArrayNumba as AppendableArray
from awkward.numba.array.indexed import IndexedArrayNumba as IndexedArray
from awkward.numba.array.indexed import SparseArrayNumba as SparseArray
from awkward.numba.array.jagged import JaggedArrayNumba as JaggedArray
from awkward.numba.array.masked import MaskedArrayNumba as MaskedArray
from awkward.numba.array.masked import BitMaskedArrayNumba as BitMaskedArray
from awkward.numba.array.masked import IndexedMaskedArrayNumba as IndexedMaskedArray
from awkward.numba.array.objects import MethodsNumba as Methods
from awkward.numba.array.objects import ObjectArrayNumba as ObjectArray
from awkward.numba.array.objects import StringArrayNumba as StringArray
from awkward.numba.array.table import TableNumba as Table
from awkward.numba.array.union import UnionArrayNumba as UnionArray
from awkward.numba.array.virtual import VirtualArrayNumba as VirtualArray

import awkward.generate
def fromiter(iterable, awkwardlib=None, **options):
    if awkwardlib is None:
        awkwardlib = "awkward.numba"
    return awkward.generate.fromiter(iterable, awkwardlib=awkwardlib, **options)

# convenient access to the version number
from awkward.version import __version__

__all__ = ["numpy", "AwkwardArray", "ChunkedArray", "AppendableArray", "IndexedArray", "SparseArray", "JaggedArray", "MaskedArray", "BitMaskedArray", "IndexedMaskedArray", "Methods", "ObjectArray", "Table", "UnionArray", "VirtualArray", "StringArray", "fromiter", "__version__"]



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import unittest

import numpy

import awkward

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_likenumpy_slices(self):
        print()

        np = numpy.array([[1, 10, 100], [2, 20, 200], [3, 30, 300]])
        aw = awkward.fromiter(np)

        assert np.tolist() == aw.tolist()
        assert np[:2].tolist() == aw[:2].tolist()
        assert np[:2, :2].tolist() == aw[:2, :2].tolist()
        assert np[:2, 2].tolist() == aw[:2, 2].tolist()
        assert np[2, :2].tolist() == aw[2, :2].tolist()
        assert np[:2, [0, 1]].tolist() == aw[:2, [0, 1]].tolist()
        assert np[[0, 1], :2].tolist() == aw[[0, 1], :2].tolist()
        assert np[:2, [0, 1, 2]].tolist() == aw[:2, [0, 1, 2]].tolist()
        assert np[[0, 1, 2], :2].tolist() == aw[[0, 1, 2], :2].tolist()
        assert np[[0, 1], [0, 1]].tolist() == aw[[0, 1], [0, 1]].tolist()
        assert np[[0, 1, 2], [0, 1, 2]].tolist() == aw[[0, 1, 2], [0, 1, 2]].tolist()
        assert np[:2, [True, False, True]].tolist() == aw[:2, [True, False, True]].tolist()
        assert np[[True, False, True], :2].tolist() == aw[[True, False, True], :2].tolist()
        assert np[[True, False, True], [True, False, True]].tolist() == aw[[True, False, True], [True, False, True]].tolist()

        np = numpy.array([[[1, 10, 100], [2, 20, 200], [3, 30, 300]], [[4, 40, 400], [5, 50, 500], [6, 60, 600]], [[7, 70, 700], [8, 80, 800], [9, 90, 900]]])
        aw = awkward.fromiter(np)

        assert np.tolist() == aw.tolist()
        assert np[:2].tolist() == aw[:2].tolist()
        assert np[:2, :2].tolist() == aw[:2, :2].tolist()
        assert np[:2, 2].tolist() == aw[:2, 2].tolist()
        assert np[2, :2].tolist() == aw[2, :2].tolist()
        assert np[:2, [0, 1]].tolist() == aw[:2, [0, 1]].tolist()
        assert np[[0, 1], :2].tolist() == aw[[0, 1], :2].tolist()
        assert np[:2, [0, 1, 2]].tolist() == aw[:2, [0, 1, 2]].tolist()
        assert np[[0, 1, 2], :2].tolist() == aw[[0, 1, 2], :2].tolist()
        assert np[[0, 1], [0, 1]].tolist() == aw[[0, 1], [0, 1]].tolist()
        assert np[[0, 1, 2], [0, 1, 2]].tolist() == aw[[0, 1, 2], [0, 1, 2]].tolist()
        assert np[:2, [True, False, True]].tolist() == aw[:2, [True, False, True]].tolist()
        assert np[[True, False, True], :2].tolist() == aw[[True, False, True], :2].tolist()
        assert np[[True, False, True], [True, False, True]].tolist() == aw[[True, False, True], [True, False, True]].tolist()

        assert np[:2, :2, 0].tolist() == aw[:2, :2, 0].tolist()
        assert np[:2, 2, 0].tolist() == aw[:2, 2, 0].tolist()
        assert np[2, :2, 0].tolist() == aw[2, :2, 0].tolist()
        assert np[:2, [0, 1], 0].tolist() == aw[:2, [0, 1], 0].tolist()
        assert np[[0, 1], :2, 0].tolist() == aw[[0, 1], :2, 0].tolist()
        assert np[:2, [0, 1, 2], 0].tolist() == aw[:2, [0, 1, 2], 0].tolist()
        assert np[[0, 1, 2], :2, 0].tolist() == aw[[0, 1, 2], :2, 0].tolist()
        assert np[[0, 1], [0, 1], 0].tolist() == aw[[0, 1], [0, 1], 0].tolist()
        assert np[[0, 1, 2], [0, 1, 2], 0].tolist() == aw[[0, 1, 2], [0, 1, 2], 0].tolist()
        assert np[:2, [True, False, True], 0].tolist() == aw[:2, [True, False, True], 0].tolist()
        assert np[[True, False, True], :2, 0].tolist() == aw[[True, False, True], :2, 0].tolist()
        assert np[[True, False, True], [True, False, True], 0].tolist() == aw[[True, False, True], [True, False, True], 0].tolist()

        assert np[:2, :2, 1].tolist() == aw[:2, :2, 1].tolist()
        assert np[:2, 2, 1].tolist() == aw[:2, 2, 1].tolist()
        assert np[2, :2, 1].tolist() == aw[2, :2, 1].tolist()
        assert np[:2, [0, 1], 1].tolist() == aw[:2, [0, 1], 1].tolist()
        assert np[[0, 1], :2, 1].tolist() == aw[[0, 1], :2, 1].tolist()
        assert np[:2, [0, 1, 2], 1].tolist() == aw[:2, [0, 1, 2], 1].tolist()
        assert np[[0, 1, 2], :2, 1].tolist() == aw[[0, 1, 2], :2, 1].tolist()
        assert np[[0, 1], [0, 1], 1].tolist() == aw[[0, 1], [0, 1], 1].tolist()
        assert np[[0, 1, 2], [0, 1, 2], 1].tolist() == aw[[0, 1, 2], [0, 1, 2], 1].tolist()
        assert np[:2, [True, False, True], 1].tolist() == aw[:2, [True, False, True], 1].tolist()
        assert np[[True, False, True], :2, 1].tolist() == aw[[True, False, True], :2, 1].tolist()
        assert np[[True, False, True], [True, False, True], 1].tolist() == aw[[True, False, True], [True, False, True], 1].tolist()
